/*
  Library for Teensy 3.6/Teensy 3.2 6.302 board Created by teaching
  staff of 6.302, a control class at MIT.
*/

#include "lib6302.h"

// Set Dac and Adc resolution, and PWM frequency for library
// Set PWM frequency, Teensy 8bit PWM: 3.6-234375, 3.2-187500!!!
// Divide by 2,4,8 or 16 for 9,10,11, or 12 bits.
// Warning 12 bit pwm in audio range for young listeners.



extern ADC *_adc;
extern volatile uint16_t _bufMax;
extern volatile uint16_t _numSamples;
extern volatile uint16_t _adcArray[];
extern volatile uint16_t _numS;

void adc0_isr(void) {
  if(_numS < _numSamples) {
    _adcArray[_numS] = (uint16_t) _adc->adc0->analogReadContinuous();
    _numS++; 
  } 
  else _adc->adc0->analogReadContinuous();
};

lib6302::lib6302(int pwmF, int dacRes, int adcRes, Stream *serialChan) {
  _PWMF = pwmF;
  _dacRes = dacRes;
  _adcRes = adcRes;
  _serChan = serialChan;
  _dacMax = float(pow(2,dacRes) - 1);
  _adcMax = float(pow(2,adcRes) - 1);
  _firstTime = true;
  _sinceToggle = 0;
  _inputString = "";
  _flipSign = 1.0;
}

void lib6302::_parseConfigStr(String inStr) {
  // Count the number of plots
  int numPlots = 0;
  for(String subStr = inStr; subStr.length() > 5;) {
    int nextP = subStr.indexOf("P~");
    subStr = subStr.substring(nextP+2);
    int endLabel = subStr.indexOf('~');
    _plotLabels[numPlots] = subStr.substring(0,endLabel);
    subStr = subStr.substring(endLabel+1);
    _plotMins[numPlots] = subStr.toFloat();
    subStr = subStr.substring(subStr.indexOf('~')+1);
    _plotMaxs[numPlots] = subStr.toFloat();
    subStr = subStr.substring(subStr.indexOf('~')+1);
    subStr = subStr.substring(subStr.indexOf('~')+1);
    int numSubPlots = subStr.toInt();
    if(numPlots == 0) {
      _plotOffsets[numPlots] = 0.0;
    } else {
      _plotOffsets[numPlots] = _plotOffsets[numPlots-1]
	+ _plotMaxs[numPlots-1] + abs(_plotMins[numPlots]);
    }
    for(int j = 1; j < numSubPlots; j++) {
      int label_break = _plotLabels[numPlots].indexOf(" ");
      _plotLabels[numPlots+1] = _plotLabels[numPlots].substring(label_break+1);
      _plotLabels[numPlots] = _plotLabels[numPlots].substring(0,label_break);
      _plotMins[numPlots+1] = _plotMins[numPlots];
      _plotMaxs[numPlots+1] = _plotMaxs[numPlots];
      _plotOffsets[numPlots+1] = _plotOffsets[numPlots];
      numPlots += 1;
    }
    numPlots += 1;
  }
  _numPlots = numPlots;
}

void lib6302::setupADC(int numAvg, boolean goFast)
{
  _adc->adc0->setAveraging(numAvg); 
  _adc->adc0->setResolution(_adcRes);
  if(goFast) {  // Go fast as possible
    _adc->adc0->setConversionSpeed(ADC_CONVERSION_SPEED::VERY_HIGH_SPEED); 
    _adc->adc0->setSamplingSpeed(ADC_SAMPLING_SPEED::VERY_HIGH_SPEED); 
  } else { // A little slower but less noisy
    _adc->adc0->setConversionSpeed(ADC_CONVERSION_SPEED::MED_SPEED); 
    _adc->adc0->setSamplingSpeed(ADC_SAMPLING_SPEED::MED_SPEED); 
  }
  _adc->adc0->stopContinuous();
  _adc->adc0->enableInterrupts(adc0_isr);
}


// In: ADC pin, num samples in a burst.
// Out: Avg val of burst, individual vals in _adcSampleArray.
int lib6302::_adcReadIn(const uint8_t pin, const uint16_t numSamples) {
  _numSamples = numSamples;  // Place where interrupt can see
  
  // Wait until the samples are collected
  _adc->adc0->wait_for_cal();
  _numS = 0;
  _adc->adc0->startContinuous(pin);
  while(_numS < _numSamples) {}  // Wait to get numSamples.
  _adc->adc0->stopContinuous();

  // Average Samples
  uint32_t sumSample = 0;
  for(int i = 0; i < _numSamples; i++) sumSample += _adcArray[i];
  return((int) (sumSample/_numSamples));
}

// Scales 0->adcMax to 0.0->1.0.
float lib6302::adcGetAvg(const uint8_t pin) {
  float avgADC = _adcReadIn(pin,1)/_adcMax;
  return(avgADC);
}

// Scales 0->adcMax to 0.0->1.0.
float lib6302::adcGetAvg(const uint8_t pin, uint16_t &numSamples) {
  uint16_t numSamplesMax = min(_bufMax,numSamples);
  numSamples = numSamplesMax;
  float avgADC = _adcReadIn(pin,numSamplesMax)/_adcMax;
  return(avgADC);
}
  
// Gets numSamples, scales 0->adcMax to 0.0->1.0.
float lib6302::adcGetBuf(const uint8_t pin, uint16_t &numSamples, float sampBuf[]) {
  uint16_t numSamplesMax = min(_bufMax,numSamples);
  numSamples = numSamplesMax;
  float avgADC = _adcReadIn(pin,numSamplesMax)/_adcMax;
  for(int i = 0; i < numSamplesMax; i++) sampBuf[i] = (_adcArray[i])/_adcMax;
  return(float(avgADC));
}

// Setup the 6302 library.
void lib6302::setup(String config_message, float togglePeriod) {

  // Initialize Serial Channel
  _serChan->println(" ");
  while(!_serChan) {}
  _serChan->println(" ");
  _serChan->setTimeout(2);
  
  _inputString = "";
  _firstTime = true;
  _config_message = "\fB~" + config_message + "\r";
  _togglePeriod = int(1.0e6*togglePeriod);

  // Set up loop monitor pin
  pinMode(loopMonitor, OUTPUT);

  // Set up inputs
  pinMode(A0, INPUT);
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);
  pinMode(A3, INPUT);
  pinMode(A4, INPUT);
  pinMode(A5, INPUT);
  pinMode(A6, INPUT);
  pinMode(A7, INPUT);
  pinMode(A8, INPUT);
  pinMode(A9, INPUT);
  pinMode(DAC0,OUTPUT);
  pinMode(DAC1,OUTPUT);

  // Set up outputs.
  pinMode(loopMonitor, OUTPUT);
  pinMode(nSleepPWM, OUTPUT);
  digitalWrite(nSleepPWM, LOW);
  analogWriteResolution(_dacRes); 
  analogWriteFrequency(pwmA1,_PWMF); 
  analogWriteFrequency(pwmA2,_PWMF); 
  analogWriteFrequency(pwmB1,_PWMF); 
  analogWriteFrequency(pwmB2,_PWMF); 
  analogWriteFrequency(pwmA1_T,_PWMF); 
  analogWriteFrequency(pwmA2_T,_PWMF); 
  analogWriteFrequency(pwmB1_T,_PWMF); 
  analogWriteFrequency(pwmB2_T,_PWMF);

  analogWrite(DAC0, _dacMax/2);  // half way.
  analogWrite(DAC1, _dacMax/2);  // half way.
  
  // Turn on and off loop monitor,diagnostic
  digitalWrite(loopMonitor,HIGH);
  delay(10);
  digitalWrite(loopMonitor,LOW);
  delay(10);
  digitalWrite(loopMonitor,HIGH);

}

// Turn on the 6302 PWM drivers
void lib6302::startPWM(boolean parallelDrive) {
  
  digitalWrite(nSleepPWM, LOW);
  for(int i = 0; i < 3; i++) {
    if(parallelDrive) {
      pinMode(pwmA1_T, INPUT);
      pinMode(pwmA2_T, INPUT);
      pinMode(pwmA1, INPUT);
      pinMode(pwmA2, INPUT);
    }
    else {
      analogWrite(pwmA1,_dacMax); 
      analogWrite(pwmA2,_dacMax);
      analogWrite(pwmA1_T,_dacMax); 
      analogWrite(pwmA2_T,_dacMax);
    }
    analogWrite(pwmB1,_dacMax); 
    analogWrite(pwmB2,_dacMax); 
    analogWrite(pwmB1_T,_dacMax); 
    analogWrite(pwmB2_T,_dacMax);

    if(i==0) digitalWrite(nSleepPWM, HIGH);
  }
}


void lib6302::_processRcvString(String inStr, float rcvBuf[]) {
  //Serial.println("got here");
  //Serial.println(inStr);
  char St = inStr.charAt(0);
  inStr.remove(0,1);
  inStr.remove(0,1);

  float val = inStr.toFloat();
  switch (St) {
    case '0':
      rcvBuf[0] = val;
      //Serial.println(val);
      break;
    case '1':
      rcvBuf[1] = val;
      break;
    case '2': 
      rcvBuf[2] = val;
      break;
    case '3':
      rcvBuf[3] = val;      
      break;  
    case '4':
      rcvBuf[4] = val;      
      break;  
    case '5':
      rcvBuf[5] = val;      
      break;
    case '6':
      rcvBuf[6] = val;      
      break;
    case '~':
      _firstTime = true;
      break;
    default:
    break;  
  }
}

// Load the serial output buffer.
void lib6302::sendStatus(int numStats, float stats[])
{
  if(_serChan == &Serial) { // Using the Arduino plotter
    if(_firstStatus == true) { // Labels.
      // Set plot mins and maxes
      //float minPlot = _plotOffsets[0]+_plotMins[0];
      //float maxPlot = _plotOffsets[mInd]+_plotMaxs[mInd];
      //_serChan->println("Min:" + String(minPlot) + ",Max:" + String(maxPlot));
      //Print the labels

      float v = 1000.0*(_plotOffsets[0]+stats[0]);
      // Tries to clear out plotter.
      for(int i = 0; i < 5; i++) {
        _serChan->print(":");
	_serChan->println(v);
      }
      delay(100);

      // Print labels
      _serChan->print(_plotLabels[0]+":");
      for(int i = 1; i < _numPlots; i++) {
	_serChan->print("," + _plotLabels[i] + ":");
      }
      _serChan->println();
      delay(100);
      
      // Do this only once!
      _firstStatus = false;
    } else {
      _serChan->print(1000.0*(_plotOffsets[0]+stats[0]));
      for(int i=1; i < _numPlots; i++) {
	_serChan->print(",");
	_serChan->print(1000.0*(_plotOffsets[i]+stats[i]));
      }
      _serChan->println();
    }
  }
  else {  // Not the Teensyduino plotter
    int n = 0;
    // Start Byte.
    _sendBuf[n++] = '\f';
    _sendBuf[n++] = 'R';

    // status data
    int dataBytes = numStats*sizeof(stats[0]);
    memcpy(&(_sendBuf[n]),&(stats[0]),dataBytes);
    n += dataBytes;

    // CR to end status buffer.
    _sendBuf[n++] = '\n';
  
    // write to the serial line.
    _serChan->write(_sendBuf,n);
  }
}

// Function to access the first time flag
boolean lib6302::isFirst()
{
  return(_firstTime);
}

// Function to access the first time flag
void lib6302::setFirst(boolean val)
{
  _firstTime = val;
}

float lib6302::getFlip()
{
  return(_flipSign);
}

float lib6302::getSlider(int whichVal)
{
  return(_rcvBuf[whichVal]);
}

float lib6302::getDacMax()
{
  return(_dacMax);
}

float lib6302::getAdcMax()
{
  return(_adcMax);
}

// Initializes the loop if this is the first time, or if reconnect sent
// from GUI.  Otherwise, just look for serial event.
void lib6302::startloop() {
  if (_firstTime) { // If first time, send config string
    // Send the configuration info.
    if(_serChan == &Serial) { // Using the Arduino plotter
      _parseConfigStr(_config_message);
      _firstTime = false;
    } else {
      _inputString = "";
      _firstTime = false;
      // Flush Transmitter
      _serChan->flush();
      // Clear out rcvr.
      while(_serChan->available()) {
        _serChan->read(); 
      }
    }
    for(float timeOut = 0; !_serChan->available() && timeOut<1e6; timeOut++);
    while(_serChan->available() > 0) _serChan->read(); // Clr Rcv
    
  } 
  else { // Check for data from GUI
    while (_serChan->available()) {
      char inChar = (char)_serChan->read();
      _inputString += inChar;
      //Serial.println(inChar);
      if (inChar == '\n') { // newline terminates line.
        _processRcvString(_inputString,_rcvBuf);
        _inputString = "";
        break;
      } else if(_inputString.length() > 50) { // Comm failure.
        _inputString = "";
        break;
      }
    }

    // Flip the monitor pin every loop
    _monitorSwitch = !_monitorSwitch;
    digitalWrite(loopMonitor, _monitorSwitch);

    // Flip the sign if its been a toggle period.
    if(int(_sinceToggle) > _togglePeriod) {
      _flipSign *= -1.0; 
      _sinceToggle = 0;
    }

  }
  /*
  for(int i = 0; i < _numPlots; i++) {
    Serial.print(_plotLabels[i]); Serial.print(", ");
    Serial.print(_plotMins[i]); Serial.print(", ");
    Serial.print(_plotMaxs[i]);  Serial.print(", ");
    Serial.println(_plotOffsets[i]);
  }
  */
}

// Write motor commands to hbridge.
// normalized (0->1) to sign-flipped hbridge, clipped (0, dMax).
float lib6302::hbridgeWrite(float normV, int pin) {
  float normVclip = min(max(normV,0.0),1.0);
  int unclip = int(normVclip*float(_dacMax));
  analogWrite(pin, _dacMax - unclip);
  return(normVclip);
}

// normV from -1->1, for maximum negative voltage to maximum positive.
// On pinP: normV from -oLap->1-oLap maps to dMax->0, 
// On pinN: normV from -1+oLap->oLap maps to dMax->0
#define oLapF 0.00  //overlap forward and backward by 5%
float lib6302::hbridgeBipolar(float normV, int pinP, int pinN) {
  float normVp = hbridgeWrite(normV + oLapF, pinP);
  float normVm = hbridgeWrite(oLapF - normV, pinN);
  float retval = normVp;
  if (normVm > normVp) retval = -normVm;
  return(retval);
}

